/*
Copyright (C) 2005-2024 Remon Sijrier

This file is part of Traverso

Traverso is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA.

*/

#include "TInputEventDispatcher.h"

#include "TContextPointer.h"
#include "TInformUser.h"
#include "TCommand.h"
#include "TMoveCommand.h"
#include "TCommandPlugin.h"
#include "TShortCut.h"
#include "TShortCutFunction.h"
#include "Utils.h"
#include "TShortCutManager.h"
#include "TConfig.h"
#include "TTimeRef.h"
#include "Debugger.h"

#include <QMetaMethod>
#include <QKeyEvent>
#include <QWheelEvent>


/**
 * \class InputEventDispatcher
 * \brief Processes keyboard/mouse events, dispatches the result, and handles the returned Command objects
 *
        InputEngine forms, together with ViewPort, Command, ContextPointer, QObject <br />
        and Qt's Undo Framework, the framework that makes up the the Contextual <br />
        Interaction Interface, with analog type of actions, and un/redo (aka History) support.


        <b>Dispatching key facts to QObject objects</b>

        InputEngine parses the keyboard/mouse events generated by the pointed ViewPort<br />
        If the keysequence matches that of any given in the keymap file, it call's<br />
        broadcast_action(). A list of pointed QObject objects is retrieved then <br />
        from ContextPointer. This list represents all (gui) QObject objects with their <br />
        corresponding 'core' QObject objects, stacked, with the topmost gui object on top.

        For each QObject in the list, the class name is retreived, and looked up in the keymap <br />
        if for the detected key fact an object was supplied with the exact same name.<br />
        If this is the case, the function name, also given in the keymap file by the given object name <br />
        is used to call the QObject's function. If succesfull, the InputEngine will stop iterating <br />
        over the list, and start handling the returned Command object.

        If the keymap specified that the object's doesn't have a function (slot) to be called, but instead<br />
        uses a TCommandPlugin, the list of loaded CommandPlugins is searched to find a match for the <br />
        plugin name supplied in the keymap file, if there is a match, the Plugin is used to create <br />
        the Command object, and the same routine is used to handle this Command object.

        If the Command object returned no error during the handling, it'll be placed on it's <br />
        historystack. If no historystack was available, it's do_action() will be called, and <br />
        deleted afterwards.


 *	\sa Command, ContextPointer, ViewPort, TCommandPlugin
 */


TInputEventDispatcher& ied()
{
    static TInputEventDispatcher inputEventDispatcher;
    return inputEventDispatcher;
}

TInputEventDispatcher::TInputEventDispatcher()
{
    PENTERCONS;
    m_contextPointer = &cpointer();

    m_modifierKeys << Qt::Key_Shift << Qt::Key_Control << Qt::Key_Alt << Qt::Key_Meta;

    connect(&m_holdKeyRepeatTimer, SIGNAL(timeout()), this, SLOT(process_hold_modifier_keys()));
}

TInputEventDispatcher::~ TInputEventDispatcher( )
    = default;

int TInputEventDispatcher::dispatch_shortcut_from_contextmenu(TShortCutFunction* function)
{
    PENTER2;
    Q_ASSERT(m_shortCutManager);

    m_contextPointer->restore_global_mouse_pos_after_context_menu_dispatch();

    QStringList keys = function->get_keys();
    if (!keys.size()) {
        return -1;
    }

    TShortCut* shortCut = m_shortCutManager->get_shortcut_for_key(keys.first());

    if (! shortCut) {
        //		PERROR("ContextMenu keySequence doesn't apply to any InputEngine knows off!! (%s)", QS_C(keys.first()));
        return -1;
    }

    m_activeModifierKeys.append(function->get_modifier_keys());

    if (function->get_command_name() == QString("RejectHoldCommand")) {
        process_press_event(Qt::Key_Escape);
        return 1;
    }

    if (function->get_command_name() == QString("AcceptHoldCommand")) {
        process_press_event(Qt::Key_Enter);
        return 1;
    }

    dispatch_shortcut(shortCut, true);

    m_activeModifierKeys.clear();

    return 1;
}


int TInputEventDispatcher::dispatch_shortcut(TShortCut* shortCut, bool fromContextMenu)
{
    PENTER2;
    Q_ASSERT(m_shortCutManager);

    PMESG("Dispatching key %d", shortCut->get_key_value());

    QList<QObject* > contextItemsList = fromContextMenu ? m_contextPointer->get_contextmenu_items() : m_contextPointer->get_context_items();

    // If we're holding the first item we want to dispatch to is the holding command object
    if (m_holdingCommand) {
        contextItemsList.prepend(m_holdingCommand);
    }

    for (QObject * contextItem : contextItemsList) {

        m_dispatchResult = BroadcastResult::RESULT_NOT_SET;

        if (!contextItem) {
            PERROR("no item in cpointer()'s context item list ??");
            continue;
        }

        const QMetaObject* metaObject = contextItem->metaObject();
        TShortCutFunction* shortCutFunction = find_shortcut_function_for_metaobject(metaObject, shortCut);

        if (! shortCutFunction ) {
            PMESG("No data found for object %s", metaObject->className());
            continue;
        }

        PMESG("Function found for metaobject %s", metaObject->className());
        PMESG("setting slotsignature to %s", QS_C(shortCutFunction->get_slot_signature()));
        PMESG("setting pluginname to %s", QS_C(shortCutFunction->get_plugin_name()));
        PMESG("setting commandname to %s", shortCutFunction->get_command_name());

        TCommand* command = nullptr;

        if (is_holding()) {
            // If we're holding we either only dispatch to the hold command object
            if (contextItem == m_holdingCommand) {
                PMESG("Dispatching to holdcommand %s", m_holdingCommand->metaObject()->className());
                if (shortCutFunction->dispatch(contextItem)) {
                    // only now we know which object this hold modifier key was dispatched on.
                    // the process_hold_modifier_keys() only knows about the corresonding ieaction
                    // next time it'll be called, autorepeat interval of the object + keysequence will
                    // be used!
                    shortCut->set_autorepeat_interval(shortCutFunction->get_autorepeat_interval());
                    shortCut->set_autorepeat_start_delay(shortCutFunction->get_autorepeat_start_delay());

                    if (shortCutFunction->uses_autorepeat()) {
                        m_holdKeyRepeatTimer.start(10);
                    }

                    return 1;
                }
                // or if that didn't apply to a shortcut function that has registered itself
                // to be always safe to be dispached to
            } else if (shortCutFunction->always_safe_to_dispatch()) {
                shortCutFunction->dispatch(contextItem);
            }

            return 1;
        } else {
            if (shortCutFunction->plugin_based()) {
                command = create_command_from_plugin(contextItem, shortCutFunction);
            } else  {
                command = shortCutFunction->dispatch_with_command_return(contextItem);
            }
        }

        // Let's see if the invoked object used either succes(), failure() or did_not_implement()
        // return functions, so we can detect to either return happily, the action was succesfull
        // but no command object needed to be returned, the action was not succesfull, and we
        // don't want to try lower level context items or the action was succesfull but we'd like
        // to give a lower level object precedence over the current one.
        switch (m_dispatchResult) {
        case SUCCESS:
            Q_ASSERT(!command);
            PMESG("Broadcast Result indicates succes, but no returned Command object");
            reset();
            return 1;
        case FAILURE:
            Q_ASSERT(!command);
            PMESG("Broadcast Result indicates failure, and doesn't want lower level items to be processed");
            reset();
            return 0;
        case DIDNOTIMPLEMENT:
            Q_ASSERT(!command);
            PMESG("Broadcast Result indicates succes, but didn't want to perform it's action,"
                  "so we continue traversing the objects list");
            continue;
        default:
            PMESG("Unkown dispatch result");
        }


        if (!command) {
            PMESG("No returned command object");
            break;
        } else {
            command->set_context_pointer(m_contextPointer);
        }

        if (command->is_hold_command()) {
            if (command->begin_hold() != -1) {

                command->set_valid(true);
                command->set_cursor_shape(shortCutFunction->uses_x(), shortCutFunction->uses_y());

                if (has_collected_number()) {
                    command->process_collected_number(get_collected_number());
                    set_numerical_input("");
                }

                m_holdingCommand = command;
                m_moveCommand = qobject_cast<TMoveCommand*>(m_holdingCommand);

                if (m_moveCommand) {
                    m_moveCommand->TMoveCommand::begin_hold();
                }

                m_holdEventKeyValue = shortCut->get_key_value();

                set_holding(true);

                m_enterFinishesHold = config().get_property("InputEventDispatcher", "EnterFinishesHold", false).toBool();

                // Override use preference for enter finishes hold behavior when the event started from a context menu
                // in which case the only way to accept the hold action is with enter
                if (fromContextMenu && command->supportsEnterFinishesHold()) {
                    m_enterFinishesHold = true;
                }

                if (!command->supportsEnterFinishesHold()) {
                    m_enterFinishesHold = false;
                }

                // If the same key to start the hold command is also used for auto repeat dispatching
                // on itself we have to simulate the same key press and start the auto repeat timer
                // Use case: arrow key starts navigation hold command, while holding that key it starts
                // to use auto repeat to call the holds command slot to navigate to the next item etc.
                if (shortCutFunction->auto_repeats_on_itself()) {
                    PMESG("Function uses autorepeat");
                    process_press_event(shortCut->get_key_value());
                    m_holdKeyRepeatTimer.start(10);
                }

                bool showCursorShortCutHelp = config().get_property("ShortCuts", "ShowCursorHelp", false).toBool();
                if (showCursorShortCutHelp && (fromContextMenu || m_enterFinishesHold)) {
                    int key = shortCut->get_key_value();
                    QString keyString;
                    if (key == Qt::LeftButton) {
                        keyString = tr("Mouse Button Left");
                    } else {
                        keyString = QKeySequence(shortCut->get_key_value()).toString();
                    }
                    m_contextPointer->set_canvas_cursor_text(tr("Press %1 or Enter to accept, Esc to cancel, Q for more").arg(keyString));
                }

                return 1;
            } else {
                PWARN("hold action begin_hold() returned -1");
                // OOPSSS, something went wrong when making the Command
                // set following stuff to zero to make finish_hold do nothing
                delete command;
                command = nullptr;
                set_holding( false );
                return 0;
            }
        }
        else {
            TCommand::process_command(command);
            return 1;
        }

        // never reached
        PERROR("ShortCut processing fallthrough detected, shortcut dispatching programming error");
    }

    return 1;
}

TShortCutFunction *TInputEventDispatcher::find_shortcut_function_for_metaobject(const QMetaObject *metaObject, TShortCut *shortCut)
{
    // traverse upwards till no more superclasses are found
    // this supports inheritance on QObjects.
    while (metaObject)
    {
        QList<TShortCutFunction*> functions = shortCut->get_shortcut_functions_for_metaobject(metaObject);

        for(TShortCutFunction* function : functions) {

            if (m_activeModifierKeys.size()) {
                if (modifierKeysMatch(m_activeModifierKeys, function->get_modifier_keys())) {
                    PMESG("found match in objectUsingModierKeys");
                    return function;
                } else {
                    PMESG("m_activeModifierKeys doesn't contain code %d", shortCut->get_key_value());
                }
            } else {
                if (function->get_modifier_keys().isEmpty()) {
                    PMESG("found match in obects NOT using modifier keys");
                    return function;
                }
            }
        }

        metaObject = metaObject->superClass();
    }

    return nullptr;
}

TCommand *TInputEventDispatcher::create_command_from_plugin(QObject* contextItem, TShortCutFunction* shortCutFunction)
{
    const QString &pluginname = shortCutFunction->get_plugin_name();
    TCommandPlugin* plug = m_shortCutManager->get_command_plugin(pluginname);
    if (!plug) {
        tInformUser().critical(tr("Command Plugin %1 not found!").arg(pluginname));
        return nullptr;
    }

    const QString& commandName = shortCutFunction->get_command_name();
    if (plug->implements(commandName)) {
        PMESG("InputEventDispatcher:: Using plugin %s for command %s", QS_C(pluginname), shortCutFunction->get_command_name());
        return plug->create(contextItem, commandName, shortCutFunction->get_arguments());
    } else {
        tInformUser().critical(tr("Plugin %1 doesn't implement Command %2").arg(pluginname, commandName));
    }
    return nullptr;
}

TCommand* TInputEventDispatcher::succes()
{
    m_dispatchResult = SUCCESS;
    return nullptr;
}

TCommand* TInputEventDispatcher::failure()
{
    m_dispatchResult = FAILURE;
    return nullptr;
}

TCommand* TInputEventDispatcher::did_not_implement()
{
    m_dispatchResult = DIDNOTIMPLEMENT;
    return nullptr;
}

void TInputEventDispatcher::jog()
{
    PENTER3;

    if (!m_isHolding) {
        PERROR("jog() called but not holding, invalid call to jog()");
        return;
    }

    if (!m_holdingCommand) {
        PERROR("jog() called but no holdingCommand but m_isHolding was set to true, internal state error, this is a bug!");
        return;
    }

    if (m_bypassJog) {
        QPoint diff = m_jogBypassPos - m_contextPointer->mouse_viewport_pos();
        if (diff.manhattanLength() > m_unbypassJogDistance) {
            m_bypassJog = false;
            m_holdingCommand->set_jog_bypassed(m_bypassJog);
        } else {
            return;
        }
        m_jogBypassPos = m_contextPointer->mouse_viewport_pos();
    }

    if (m_holdingCommand->jog() == 1 && m_holdingCommand->canvas_cursor_follows_mouse_cursor() && m_moveCommand) {
        m_moveCommand->TMoveCommand::jog();
        m_contextPointer->set_canvas_cursor_pos(m_contextPointer->scene_pos());
    }
}

void TInputEventDispatcher::bypass_jog_until_mouse_movements_exceeded_manhattenlength(int length)
{
    m_unbypassJogDistance = length;
    m_bypassJog = true;
    m_jogBypassPos = m_contextPointer->mouse_viewport_pos();
    if (m_holdingCommand) {
        m_holdingCommand->set_jog_bypassed(m_bypassJog);
    }
}

void TInputEventDispatcher::update_jog_bypass_pos()
{
    m_jogBypassPos = m_contextPointer->mouse_viewport_pos();
}

void TInputEventDispatcher::set_holding(bool holding)
{
    PENTER;

    m_isHolding = holding;

    if (m_isHolding) {
        emit holdStarted();
        m_contextPointer->hold_start();
    } else {
        m_contextPointer->hold_finished();
        emit holdFinished();
    }
}

void TInputEventDispatcher::reset()
{
    PENTER;
    set_holding(false);
    m_cancelHold = false;
    m_bypassJog = false;
    m_enterFinishesHold = false;

    set_numerical_input("");
}

void TInputEventDispatcher::reject_current_hold_actions()
{
    m_activeModifierKeys.clear();
    clear_hold_modifier_keys();
    // Fake an escape key press, so if a hold action was
    // running it will be canceled!
    if (is_holding()) {
        process_press_event(Qt::Key_Escape);
    }
}

void TInputEventDispatcher::catch_key_press(QKeyEvent * e )
{
    if (e->isAutoRepeat()) {
        return;
    }
    PENTER;

    process_press_event(e->key());
}

void TInputEventDispatcher::catch_key_release( QKeyEvent * e)
{
    if (e->isAutoRepeat()) {
        return;
    }
    PENTER;
    process_release_event(e->key());
}

void TInputEventDispatcher::catch_mousebutton_press( QMouseEvent * e )
{
    if (e->button() == Qt::LeftButton) {
        m_contextPointer->mouse_button_left_pressed();
    }
    process_press_event(int(e->button()));
}

void TInputEventDispatcher::catch_mousebutton_release( QMouseEvent * e )
{
    process_release_event(int(e->button()));
}

void TInputEventDispatcher::catch_scroll(QWheelEvent* e)
{
    if (e->angleDelta().y() > 0) {
        process_press_event(TShortCutManager::MouseScrollVerticalUp);
        process_release_event(TShortCutManager::MouseScrollVerticalUp);
    }
    if (e->angleDelta().y() < 0) {
        process_press_event(TShortCutManager::MouseScrollVerticalDown);
        process_release_event(TShortCutManager::MouseScrollVerticalDown);
    }
}

void TInputEventDispatcher::process_press_event(int keyValue)
{
    PENTER;
    Q_ASSERT(m_shortCutManager);

    if (keyValue == Qt::Key_Escape && is_holding())
    {
        m_cancelHold = true;
        finish_hold();
        return;
    }

    if (keyValue == m_holdEventKeyValue && is_holding() && m_holdingCommand->supportsEnterFinishesHold())
    {
        finish_hold();
        return;
    }


    if ((keyValue == Qt::Key_Return || keyValue == Qt::Key_Enter) && m_enterFinishesHold)
    {
        finish_hold();
        return;
    }

    // first check if this key is just a collected number
    if (check_number_collection(keyValue))
    {
        // another digit was collected.
        return;
    }

    if (is_modifier_keyfact(keyValue))
    {
        if (!m_activeModifierKeys.contains(keyValue))
        {
            m_activeModifierKeys.append(keyValue);
        }
        return;
    }

    TShortCut* shortCut = m_shortCutManager->get_shortcut_for_key(keyValue);

    if (m_isHolding && shortCut)
    {
        HoldModifierKey hmk;
        hmk.keycode = keyValue;
        hmk.wasExecuted = false;
        hmk.lastTimeExecuted = 0;
        hmk.shortcut = shortCut;
        m_holdModifierKeys.insert(keyValue, hmk);
        // execute the first one directly, this is needed
        // if the release event comes before the timer actually
        // fires (mouse scroll wheel does press/release events real quick
        process_hold_modifier_keys();
        return;
    }

    if (shortCut)
    {
        m_contextPointer->prepare_for_shortcut_dispatch();
        dispatch_shortcut(shortCut);
        return;
    }
}

void TInputEventDispatcher::process_release_event(int eventcode)
{

    if (is_modifier_keyfact(eventcode)) {
        m_activeModifierKeys.removeAll(eventcode);
        return;
    }

    if (m_isHolding) {
        if (m_holdModifierKeys.contains(eventcode)) {
            m_holdModifierKeys.remove(eventcode);
            if (m_holdModifierKeys.isEmpty()) {
                m_holdKeyRepeatTimer.stop();
            }
        }

        if (eventcode != m_holdEventKeyValue) {
            PMESG("release event during hold action, but NOT for holdaction itself!!");
            return;
        } else {
            if (m_enterFinishesHold)
            {
                PMESG("Only Enter or Esc keys are accepted to finish a hold command");
                return;
            }
            PMESG("release event for hold action detected!");
            finish_hold();
        }
    }
}

void TInputEventDispatcher::process_hold_modifier_keys()
{
    PENTER3;
    if (m_holdModifierKeys.empty()) {
        m_holdKeyRepeatTimer.stop();
        return;
    }

    for(HoldModifierKey &hmk : m_holdModifierKeys) {
        if (!hmk.wasExecuted) {
            hmk.wasExecuted = true;
            dispatch_shortcut(hmk.shortcut);
            hmk.lastTimeExecuted = TTimeRef::get_milliseconds_since_epoch() + hmk.shortcut->get_autorepeat_start_delay();
            continue;
        }

        trav_time_t timeDiff = (TTimeRef::get_milliseconds_since_epoch() - hmk.lastTimeExecuted);
        // if timeDiff is very close (-2 ms) to it's interval value, execute it still
        // else the next interval might be too long between the previous one.
        if ((timeDiff + 2) >= hmk.shortcut->get_autorepeat_interval()) {
            hmk.lastTimeExecuted = TTimeRef::get_milliseconds_since_epoch();
            dispatch_shortcut(hmk.shortcut);
        }
    }
}

bool TInputEventDispatcher::is_modifier_keyfact(int keyValue)
{
    return m_modifierKeys.contains(keyValue);
}

void TInputEventDispatcher::finish_hold()
{
    PENTER;
    PMESG("Finishing hold action %s", m_holdingCommand->metaObject()->className());

    m_holdEventKeyValue = NO_HOLD_EVENT;

    clear_hold_modifier_keys();

    if (m_holdingCommand->wants_cursor_position_to_be_restored())
    {
        m_contextPointer->set_canvas_cursor_pos(m_contextPointer->on_first_input_event_scene_pos());
    }

    if (m_cancelHold) {
        PMESG("Canceling this hold command");
        if (m_holdingCommand) {
            m_holdingCommand->cancel_action();
            if (m_moveCommand) {
                m_moveCommand->TMoveCommand::cancel_action();
            }
            delete m_holdingCommand;
            m_holdingCommand = nullptr;
            m_moveCommand = nullptr;
        }
    } else if (m_holdingCommand) {

        int holdFinish = m_holdingCommand->finish_hold();
        if (m_moveCommand) {
            m_moveCommand->TMoveCommand::finish_hold();
        }
        int holdprepare = -1;

        if (holdFinish > 0) {
            holdprepare = m_holdingCommand->prepare_actions();
            if (holdprepare > 0) {
                PMESG("holdingCommand->prepare_actions() returned succes!");
                m_holdingCommand->set_valid(true);
            } else {
                PMESG("holdingCommand->prepare_actions() returned <= 0, so either it failed, or nothing happened!");
                m_holdingCommand->set_valid( false );
            }
        } else {
            PMESG("holdingCommand->finish_hold() returned <= 0, so either it failed, or nothing happened!");
            m_holdingCommand->set_valid( false );
        }

        if (m_holdingCommand->push_to_history_stack() < 0) {
            if (holdprepare == 1) {
                m_holdingCommand->do_action();
            }
            delete m_holdingCommand;
        }

        m_holdingCommand = nullptr;
    }

    reset();
}

void TInputEventDispatcher::clear_hold_modifier_keys()
{
    m_holdKeyRepeatTimer.stop();
    m_holdModifierKeys.clear();
}

bool TInputEventDispatcher::check_number_collection(int eventcode)
{
    if (!m_activeModifierKeys.isEmpty())
    {
        return false;
    }

    if (((eventcode >= Qt::Key_0) && (eventcode <= Qt::Key_9)) ||
        (eventcode == Qt::Key_Comma) || (eventcode == Qt::Key_Period)) {
        set_numerical_input(m_sCollectedNumber + QChar(eventcode));
        PMESG("Collected %s so far...", QS_C(m_sCollectedNumber) ) ;
        return true;
    }
    if (eventcode == Qt::Key_Backspace) {
        if (m_sCollectedNumber.size() > 0) {
            set_numerical_input(m_sCollectedNumber.left(m_sCollectedNumber.size() - 1));
        }
        return true;
    }
    if (eventcode == Qt::Key_Minus) {
        if (m_sCollectedNumber.contains("-")) {
            set_numerical_input(m_sCollectedNumber.remove("-"));
        } else {
            set_numerical_input(m_sCollectedNumber.prepend("-"));
        }
    }
    return false;
}

bool TInputEventDispatcher::has_collected_number()
{
    return !m_sCollectedNumber.isEmpty();
}

void TInputEventDispatcher::set_numerical_input(const QString &number)
{
    m_sCollectedNumber = number;

    if (m_holdingCommand) {
        m_holdingCommand->process_collected_number(m_sCollectedNumber);
    }

    emit collectedNumberChanged();
}

bool TInputEventDispatcher::is_holding( )
{
    return m_isHolding;
}

TCommand * TInputEventDispatcher::get_holding_command() const
{
    return m_holdingCommand;
}

bool TInputEventDispatcher::modifierKeysMatch(QList<int> first, const QList<int>& second)
{
    if (first.size() != second.size())
    {
        return false;
    }

    foreach(int key, first)
    {
        if (!second.contains(key))
        {
            return false;
        }
    }

    return true;
}
