/*
Copyright (C) 2005-2008 Remon Sijrier

This file is part of Traverso

Traverso is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA.

*/

#include "TAudioBusVUMonitorWidget.h"

#include "TBusTrack.h"
#include "TVUMeterWidget.h"

#include <TProjectManager.h>
#include <TProject.h>
#include <TSheet.h>
#include <TThemer.h>
#include "TAudioDevice.h"
#include <AudioBus.h>
#include <QHBoxLayout>
#include <QMenu>
#include <QKeyEvent>
#include <QMouseEvent>



#include "Debugger.h"


TAudioBusVUMonitorWidget::TAudioBusVUMonitorWidget(QWidget* parent)
    : QWidget( parent)
{
    PENTERCONS;

    setAutoFillBackground(false);
    m_masterOutMeter = nullptr;
    m_projectMaster = nullptr;
    m_sheet = nullptr;
    m_layout = nullptr;

    create_vu_meters();

    m_menu = nullptr;

    connect(&audiodevice(), SIGNAL(driverParamsChanged()), this, SLOT(create_vu_meters()));
    connect(&pm(), &TProjectManager::projectLoaded, this, &TAudioBusVUMonitorWidget::set_project);
}


TAudioBusVUMonitorWidget::~TAudioBusVUMonitorWidget()
{
    PENTERDES;
}

QSize TAudioBusVUMonitorWidget::sizeHint() const
{
    int width = 0;
    foreach(QWidget* widget, outMeters) {
        if (! widget->isHidden()) {
            width += widget->width();
        }
    }
    return QSize(width, 140);
}

QSize TAudioBusVUMonitorWidget::minimumSizeHint() const
{
    return QSize(50, 50);
}

void TAudioBusVUMonitorWidget::create_vu_meters( )
{
    PENTER;

    if (!pm().get_project()) {
        return;
    }

    if (m_layout) delete m_layout;

    m_layout = new QHBoxLayout(this);
    m_layout->setContentsMargins(0, 0, 0, 0);
    setLayout(m_layout);

    while( ! inMeters.isEmpty() ) {
        TVUMeterWidget* meter = inMeters.takeFirst();
        m_layout->removeWidget( meter );
        delete meter;
    }

    while ( ! outMeters.isEmpty() ) {
        TVUMeterWidget* meter = outMeters.takeFirst();
        m_layout->removeWidget( meter );
        delete meter;
    }

    m_layout->addStretch(1);


    //FIXME: restarting the Driver crashes here!!!!!

    QStringList list = pm().get_project()->get_capture_buses_names();
    //        foreach(QString name, list) {
    //                AudioBus* bus = pm().get_project()->get_capture_bus(name);
    //                if (!bus) continue;
    //		VUMeter* meter = new VUMeter( this, bus );
    //		connect(bus, SIGNAL(monitoringPeaksStarted()), meter, SLOT(peak_monitoring_started()));
    //		connect(bus, SIGNAL(monitoringPeaksStopped()), meter, SLOT(peak_monitoring_stopped()));
    //                m_layout->addWidget(meter);
    //		inMeters.append(meter);
    //		meter->hide();
    //	}

    // TODO: show playback VU's on demand or ??
    list = pm().get_project()->get_playback_buses_names();
    if (list.size()) {
        //                VUMeter* meter = new VUMeter( this, audiodevice().get_playback_bus(list.at(0)));
        //		layout->addWidget(meter);
        //		outMeters.append(meter);
    }

    m_layout->addSpacing(4);

    if (m_masterOutMeter) {
        m_layout->removeWidget(m_masterOutMeter);
        delete m_masterOutMeter;
        m_masterOutMeter = nullptr;
    }
    if (m_projectMaster) {
        m_layout->removeWidget(m_projectMaster);
        delete m_projectMaster;
        m_projectMaster = nullptr;
    }

    if (m_sheet) {
        AudioBus* bus = m_sheet->get_master_out_bus_track()->get_process_bus();
        m_masterOutMeter = new TVUMeterWidget(this, bus);
        m_layout->addWidget(m_masterOutMeter);
        bus = m_sheet->get_project()->get_master_out_bus_track()->get_process_bus();
        m_projectMaster = new TVUMeterWidget(this, bus);
        m_layout->addWidget(m_projectMaster);
    }

}

void TAudioBusVUMonitorWidget::set_project(TProject * project)
{
    Q_UNUSED(project);

    if (project) {
        connect(project, &TProject::currentSessionChanged, this, &TAudioBusVUMonitorWidget::set_session);
    }
    m_sheet = 0;
}

void TAudioBusVUMonitorWidget::set_session(TSession* session)
{
    TProject* project = qobject_cast<TProject*>(session);
    if (project) {
        return;
    }

    if (session) {
        m_sheet = qobject_cast<TSheet*>(session);
        if (!m_sheet) {
            m_sheet = qobject_cast<TSheet*>(session->get_parent_session());
        }
    }

    create_vu_meters();
}

void TAudioBusVUMonitorWidget::keyPressEvent(QKeyEvent * event)
{
    if (event->isAutoRepeat()) {
        return;
    } else if (event->key() == Qt::Key_R) {
        reset_vu_meters();
    } else {
        QWidget::keyPressEvent(event);
    }
}

void TAudioBusVUMonitorWidget::mousePressEvent(QMouseEvent * event)
{
    if (event->button() == Qt::RightButton) {
        show_menu();
    }
}

void TAudioBusVUMonitorWidget::reset_vu_meters()
{
    foreach(TVUMeterWidget* meter, inMeters) {
        meter->reset();
    }
    foreach(TVUMeterWidget* meter, outMeters) {
        meter->reset();
    }
    if (m_masterOutMeter) {
        m_masterOutMeter->reset();
    }
    if (m_projectMaster) {
        m_projectMaster->reset();
    }
}

void TAudioBusVUMonitorWidget::show_menu()
{
    if (!m_menu) {
        m_menu = new QMenu(this);
        QAction* action = m_menu->addAction("Bus Monitor");
        QFont font(themer()->get_font("ContextMenu:fontscale:actions"));
        font.setBold(true);
        action->setFont(font);
        action->setEnabled(false);
        m_menu->addSeparator();
        action = m_menu->addAction("Reset VU's");
        action->setShortcut(QKeySequence("R"));
#if QT_VERSION >= QT_VERSION_CHECK(5, 10, 0)
        action->setShortcutVisibleInContextMenu(true);
#endif
        connect(action, SIGNAL(triggered(bool)), this, SLOT(reset_vu_meters()));
    }

    m_menu->exec(QCursor::pos());
}

void TAudioBusVUMonitorWidget::enterEvent(QEnterEvent *)
{
    setFocus();
}

